/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.mobicents.servlet.sip.pbx.proxy;

import java.io.IOException;
import java.util.List;

import javax.servlet.sip.Proxy;
import javax.servlet.sip.ServletParseException;
import javax.servlet.sip.SipFactory;
import javax.servlet.sip.SipServletRequest;
import javax.servlet.sip.SipServletResponse;
import javax.servlet.sip.SipURI;
import javax.servlet.sip.TooManyHopsException;
import javax.servlet.sip.URI;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.mobicents.servlet.sip.pbx.address.AddressHelper;
import org.mobicents.servlet.sip.pbx.address.AddressService;
import org.mobicents.servlet.sip.pbx.location.Binding;
import org.mobicents.servlet.sip.pbx.location.LocationService;
import org.mobicents.servlet.sip.pbx.routing.Route;
import org.mobicents.servlet.sip.pbx.routing.RouteService;

/**
 * @author Thomas Leseney
 */
public class ProxyService {

    private static Log logger = LogFactory.getLog(ProxyService.class);
    private SipFactory sipFactory;
    private LocationService locationService;
    private AddressService addressService;
    private RouteService routeService;

    public void setLocationService(LocationService locationService) {
        this.locationService = locationService;
    }

    public void setSipFactory(SipFactory sipFactory) {
        this.sipFactory = sipFactory;
    }

    public void setAddressService(AddressService addressService) {
        this.addressService = addressService;
    }

    public void setRouteService(RouteService routeService) {
        this.routeService = routeService;
    }

    public void doRequest(SipServletRequest request) throws IOException {

        if (!request.isInitial()) {
            return;
        }

        try {
            
            if (AddressHelper.isStation(request.getRequestURI())) {
                String aor = addressService.getAddressOfRecord(request.getRequestURI());
                doStationToStation(request, aor);

            } else if (AddressHelper.isDID(request.getRequestURI())) {
                doPSTN(request);

            } else {
                request.createResponse(SipServletResponse.SC_NOT_FOUND,
                        "Invalid destination number");
            }

        } catch (Exception e) {
            logger.error("Exception while proxying request", e);
            request.createResponse(SipServletResponse.SC_SERVER_INTERNAL_ERROR,
                    e.getMessage()).send();
        }
    }

    private void doStationToStation(SipServletRequest request, String aor)
            throws IOException, ServletParseException, TooManyHopsException {
        logger.debug("Initial request (" + request.getMethod() + ") for AOR: " + aor);
        List<Binding> bindings = locationService.getBindings(aor);
        if (bindings != null && bindings.size() > 0) {
            forward(request, createTargetURI(bindings.get(0)));
        } else {
            logger.debug("No binding found for AOR: " + aor);
            request.createResponse(SipServletResponse.SC_NOT_FOUND).send();
        }
    }

    private void forward(SipServletRequest request, URI target)
            throws TooManyHopsException {
        logger.debug("Proxying request to: " + target);
        Proxy proxy = request.getProxy();
        proxy.setRecordRoute(true);
        proxy.proxyTo(target);
    }

    private URI createTargetURI(Binding binding) throws ServletParseException {
        return sipFactory.createURI(binding.getContact());
    }

    private void doPSTN(SipServletRequest request) throws ServletParseException, 
            TooManyHopsException, IOException {

        logger.debug("Processing PSTN call");

        Route route = routeService.getRoute((SipURI)request.getRequestURI());
        if (route != null) {
            forward(request, sipFactory.createURI("sip:" + route.getGateway().getAddress()));
            
        } else {
            request.createResponse(SipServletResponse.SC_NOT_FOUND,
                    "Route not found for " + request.getRequestURI()).send();
        }
    }
}
